from dataclasses import dataclass

from dbt_common.exceptions import DbtRuntimeError


def compare_versions(v1: str, v2: str) -> int:
    v1_parts = v1.split('.')
    v2_parts = v2.split('.')
    for part1, part2 in zip(v1_parts, v2_parts):
        try:
            if int(part1) != int(part2):
                return 1 if int(part1) > int(part2) else -1
        except ValueError:
            raise DbtRuntimeError("Version must consist of only numbers separated by '.'")
    return 0



from dataclasses import dataclass
from typing import Any, Dict, Optional, Union

from dbt.adapters.contracts.connection import Credentials
from dbt_common.exceptions import DbtRuntimeError


@dataclass
class ClickHouseCredentials(Credentials):
    """
    ClickHouse connection credentials data class.
    """

    driver: Optional[str] = None
    host: str = 'localhost'
    port: Optional[int] = None
    user: Optional[str] = 'default'
    retries: int = 1
    database: Optional[str] = ''
    schema: Optional[str] = 'default'
    password: str = ''
    cluster: Optional[str] = None
    database_engine: Optional[str] = None
    cluster_mode: bool = False
    secure: bool = False
    verify: bool = True
    connect_timeout: int = 10
    send_receive_timeout: int = 300
    sync_request_timeout: int = 5
    compress_block_size: int = 1048576
    compression: str = ''
    check_exchange: bool = True
    custom_settings: Optional[Dict[str, Any]] = None
    use_lw_deletes: bool = False
    local_suffix: str = 'local'
    local_db_prefix: str = ''
    allow_automatic_deduplication: bool = False
    tcp_keepalive: Union[bool, tuple[int, int, int], list[int]] = False

    @property
    def type(self):
        return 'clickhouse'

    @property
    def unique_field(self):
        return self.host

    def __post_init__(self):
        if self.database and self.database != self.schema:
            raise DbtRuntimeError(
                f'    schema: {self.schema} \n'
                f'    database: {self.database} \n'
                f'    cluster: {self.cluster} \n'
                f'On Clickhouse, database must be omitted or have the same value as'
                f' schema.'
            )
        self.database = ''

        # clickhouse_driver expects tcp_keepalive to be a tuple if it's not a boolean
        if isinstance(self.tcp_keepalive, list):
            self.tcp_keepalive = tuple(self.tcp_keepalive)

    def _connection_keys(self):
        return (
            'driver',
            'host',
            'port',
            'user',
            'schema',
            'retries',
            'database_engine',
            'cluster_mode',
            'secure',
            'verify',
            'connect_timeout',
            'send_receive_timeout',
            'sync_request_timeout',
            'compress_block_size',
            'compression',
            'check_exchange',
            'custom_settings',
            'use_lw_deletes',
            'allow_automatic_deduplication',
            'tcp_keepalive',
        )



BS = '\\'
must_escape = (BS, '\'', '`')


def quote_identifier(identifier: str):
    first_char = identifier[0]
    if first_char in ('`', '"') and identifier[-1] == first_char:
        # Identifier is already quoted, assume that it's valid
        return identifier
    return f'`{escape_str(identifier)}`'


def escape_str(value: str):
    return ''.join(f'{BS}{c}' if c in must_escape else c for c in value)



from typing import List

import clickhouse_driver
import pkg_resources
from clickhouse_driver.errors import NetworkError, SocketTimeoutError
from dbt.adapters.__about__ import version as dbt_adapters_version
from dbt_common.exceptions import DbtDatabaseError

from dbt.adapters.clickhouse import ClickHouseColumn, ClickHouseCredentials
from dbt.adapters.clickhouse.__version__ import version as dbt_clickhouse_version
from dbt.adapters.clickhouse.dbclient import ChClientWrapper, ChRetryableException
from dbt.adapters.clickhouse.logger import logger

try:
    driver_version = pkg_resources.get_distribution('clickhouse-driver').version
except pkg_resources.ResolutionError:
    driver_version = 'unknown'


class ChNativeClient(ChClientWrapper):
    def query(self, sql, **kwargs):
        try:
            return NativeClientResult(self._client.execute(sql, with_column_types=True, **kwargs))
        except clickhouse_driver.errors.Error as ex:
            raise DbtDatabaseError(str(ex).strip()) from ex

    def command(self, sql, **kwargs):
        try:
            result = self._client.execute(sql, **kwargs)
            if len(result) and len(result[0]):
                return result[0][0]
        except clickhouse_driver.errors.Error as ex:
            raise DbtDatabaseError(str(ex).strip()) from ex

    def columns_in_query(self, sql: str, **kwargs) -> List[ClickHouseColumn]:
        try:
            _, columns = self._client.execute(
                f"SELECT * FROM ( \n" f"{sql} \n" f") LIMIT 0",
                with_column_types=True,
            )
            return [ClickHouseColumn.create(column[0], column[1]) for column in columns]
        except clickhouse_driver.errors.Error as ex:
            raise DbtDatabaseError(str(ex).strip()) from ex

    def get_ch_setting(self, setting_name):
        try:
            result = self._client.execute(
                f"SELECT value, readonly FROM system.settings WHERE name = '{setting_name}'"
            )
        except clickhouse_driver.errors.Error as ex:
            logger.warn('Unexpected error retrieving ClickHouse server setting', ex)
            return None
        return (result[0][0], result[0][1]) if result else (None, 0)

    def close(self):
        self._client.disconnect()

    def _create_client(self, credentials: ClickHouseCredentials):
        client = clickhouse_driver.Client(
            host=credentials.host,
            port=credentials.port,
            user=credentials.user,
            password=credentials.password,
            client_name=f'dbt-adapters/{dbt_adapters_version} dbt-clickhouse/{dbt_clickhouse_version} clickhouse-driver/{driver_version}',
            secure=credentials.secure,
            verify=credentials.verify,
            connect_timeout=credentials.connect_timeout,
            send_receive_timeout=credentials.send_receive_timeout,
            sync_request_timeout=credentials.sync_request_timeout,
            compress_block_size=credentials.compress_block_size,
            compression=False if credentials.compression == '' else credentials.compression,
            tcp_keepalive=credentials.tcp_keepalive,
            settings=self._conn_settings,
        )
        try:
            client.connection.connect()
        except (SocketTimeoutError, NetworkError) as ex:
            raise ChRetryableException(str(ex)) from ex
        return client

    def _set_client_database(self):
        # After we know the database exists, reconnect to that database if appropriate
        if self._client.connection.database != self.database:
            self._client.connection.disconnect()
            self._client.connection.database = self.database
            self._client.connection.connect()

    def _server_version(self):
        server_info = self._client.connection.server_info
        return (
            f'{server_info.version_major}.{server_info.version_minor}.{server_info.version_patch}'
        )


class NativeClientResult:
    def __init__(self, native_result):
        self.result_set = native_result[0]
        self.column_names = [col[0] for col in native_result[1]]



from dbt.adapters.base import AdapterPlugin

from dbt.adapters.clickhouse.column import ClickHouseColumn  # noqa
from dbt.adapters.clickhouse.connections import ClickHouseConnectionManager  # noqa
from dbt.adapters.clickhouse.credentials import ClickHouseCredentials
from dbt.adapters.clickhouse.impl import ClickHouseAdapter
from dbt.adapters.clickhouse.relation import ClickHouseRelation  # noqa
from dbt.include import clickhouse  # noqa

Plugin = AdapterPlugin(
    adapter=ClickHouseAdapter,
    credentials=ClickHouseCredentials,
    include_path=clickhouse.PACKAGE_PATH,
)



import re
import time
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Optional, Tuple, Union

import dbt.exceptions
from dbt.adapters.contracts.connection import AdapterResponse, Connection
from dbt.adapters.sql import SQLConnectionManager

from dbt.adapters.clickhouse.dbclient import ChRetryableException, get_db_client
from dbt.adapters.clickhouse.logger import logger

if TYPE_CHECKING:
    import agate

retryable_exceptions = [ChRetryableException]
ddl_re = re.compile(r'^\s*(CREATE|DROP|ALTER)\s', re.IGNORECASE)


class ClickHouseConnectionManager(SQLConnectionManager):
    """
    ClickHouse Connector connection manager.
    """

    TYPE = 'clickhouse'

    @contextmanager
    def exception_handler(self, sql):
        try:
            yield
        except Exception as exp:
            logger.debug('Error running SQL: {}', sql)
            if isinstance(exp, dbt.exceptions.DbtRuntimeError):
                raise
            raise dbt.exceptions.DbtRuntimeError('ClickHouse exception:  ' + str(exp)) from exp

    @classmethod
    def open(cls, connection):
        if connection.state == 'open':
            logger.debug('Connection is already open, skipping open.')
            return connection
        credentials = cls.get_credentials(connection.credentials)

        def connect():
            return get_db_client(credentials)

        return cls.retry_connection(
            connection,
            connect=connect,
            logger=logger,
            retry_limit=credentials.retries,
            retryable_exceptions=retryable_exceptions,
        )

    def cancel(self, connection):
        connection_name = connection.name
        logger.debug('Cancelling query \'{}\'', connection_name)
        connection.handle.close()
        logger.debug('Cancel query \'{}\'', connection_name)

    def release(self):
        pass  # There is no "release" type functionality in the existing ClickHouse connectors

    @classmethod
    def get_table_from_response(cls, response, column_names) -> "agate.Table":
        """
        Build agate table from response.
        :param response: ClickHouse query result
        :param column_names: Table column names
        """
        from dbt_common.clients.agate_helper import table_from_data_flat

        data = []
        for row in response:
            data.append(dict(zip(column_names, row)))

        return table_from_data_flat(data, column_names)

    def execute(
        self, sql: str, auto_begin: bool = False, fetch: bool = False, limit: Optional[int] = None
    ) -> Tuple[AdapterResponse, "agate.Table"]:
        # Don't try to fetch result of clustered DDL responses, we don't know what to do with them
        if fetch and ddl_re.match(sql):
            fetch = False

        sql = self._add_query_comment(sql)
        conn = self.get_thread_connection()
        client = conn.handle

        with self.exception_handler(sql):
            logger.debug(f'On {conn.name}: {sql}...')
            pre = time.time()
            if fetch:
                query_result = client.query(sql)
            else:
                query_result = client.command(sql)
            status = self.get_status(client)
            logger.debug(f'SQL status: {status} in {(time.time() - pre):.2f} seconds')
            if fetch:
                table = self.get_table_from_response(
                    query_result.result_set, query_result.column_names
                )
            else:
                from dbt_common.clients.agate_helper import empty_table

                table = empty_table()
            return AdapterResponse(_message=status), table

    def add_query(
        self,
        sql: str,
        auto_begin: bool = True,
        bindings: Optional[Any] = None,
        abridge_sql_log: bool = False,
    ) -> Tuple[Connection, Any]:
        sql = self._add_query_comment(sql)
        conn = self.get_thread_connection()
        client = conn.handle
        with self.exception_handler(sql):
            logger.debug(f'On {conn.name}: {sql}...')
            pre = time.time()
            client.command(sql)
            status = self.get_status(client)
            logger.debug(f'SQL status: {status} in {(time.time() - pre):0.2f} seconds')
            return conn, None

    @classmethod
    def get_credentials(cls, credentials):
        """
        Returns ClickHouse credentials
        """
        return credentials

    @classmethod
    def get_status(cls, _):
        """
        Returns connection status
        """
        return 'OK'

    @classmethod
    def get_response(cls, _):
        return 'OK'

    def begin(self):
        pass

    def commit(self):
        pass

    @classmethod
    def data_type_code_to_name(cls, type_code: Union[int, str]) -> str:
        assert isinstance(type_code, int)
        return ''



import re
from dataclasses import dataclass, field
from typing import Any, List, Literal, TypeVar

from dbt.adapters.base.column import Column
from dbt_common.exceptions import DbtRuntimeError

Self = TypeVar('Self', bound='ClickHouseColumn')


@dataclass
class ClickHouseColumn(Column):
    TYPE_LABELS = {
        'STRING': 'String',
        'TIMESTAMP': 'DateTime',
        'FLOAT': 'Float32',
        'INTEGER': 'Int32',
    }
    is_nullable: bool = False
    is_low_cardinality: bool = False
    _low_card_regex = re.compile(r'^LowCardinality\((.*)\)$')
    _nullable_regex = re.compile(r'^Nullable\((.*)\)$')
    _fix_size_regex = re.compile(r'FixedString\((.*?)\)')
    _decimal_regex = re.compile(r'Decimal\((\d+), (\d+)\)')

    def __init__(self, column: str, dtype: str) -> None:
        char_size = None
        numeric_precision = None
        numeric_scale = None

        dtype = self._inner_dtype(dtype)

        if dtype.lower().startswith('fixedstring'):
            match_sized = self._fix_size_regex.search(dtype)
            if match_sized:
                char_size = int(match_sized.group(1))

        if dtype.lower().startswith('decimal'):
            match_dec = self._decimal_regex.search(dtype)
            numeric_precision = 0
            numeric_scale = 0
            if match_dec:
                numeric_precision = int(match_dec.group(1))
                numeric_scale = int(match_dec.group(2))

        super().__init__(column, dtype, char_size, numeric_precision, numeric_scale)

    def __repr__(self) -> str:
        return f'<ClickhouseColumn {self.name} ({self.data_type}, is nullable: {self.is_nullable})>'

    @property
    def data_type(self) -> str:
        if self.is_string():
            data_t = self.string_type(self.string_size())
        elif self.is_numeric():
            data_t = self.numeric_type(self.dtype, self.numeric_precision, self.numeric_scale)
        else:
            data_t = self.dtype

        if self.is_nullable or self.is_low_cardinality:
            data_t = self.nested_type(data_t, self.is_low_cardinality, self.is_nullable)

        return data_t

    def is_string(self) -> bool:
        return self.dtype.lower() in [
            'string',
            'fixedstring',
            'longblob',
            'longtext',
            'tinytext',
            'text',
            'varchar',
            'mediumblob',
            'blob',
            'tinyblob',
            'char',
            'mediumtext',
        ] or self.dtype.lower().startswith('fixedstring')

    def is_integer(self) -> bool:
        return self.dtype.lower().startswith('int') or self.dtype.lower().startswith('uint')

    def is_numeric(self) -> bool:
        return self.dtype.lower().startswith('decimal')

    def is_float(self) -> bool:
        return self.dtype.lower().startswith('float')

    def string_size(self) -> int:
        if not self.is_string():
            raise DbtRuntimeError('Called string_size() on non-string field!')

        if not self.dtype.lower().startswith('fixedstring') or self.char_size is None:
            return 256
        else:
            return int(self.char_size)

    @classmethod
    def string_type(cls, size: int) -> str:
        return 'String'

    @classmethod
    def numeric_type(cls, dtype: str, precision: Any, scale: Any) -> str:
        return f'Decimal({precision}, {scale})'

    @classmethod
    def nested_type(cls, dtype: str, is_low_cardinality: bool, is_nullable: bool) -> str:
        template = "{}"
        if is_low_cardinality:
            template = template.format("LowCardinality({})")
        if is_nullable:
            template = template.format("Nullable({})")
        return template.format(dtype)

    def literal(self, value):
        return f'to{self.dtype}({value})'

    def can_expand_to(self, other_column: 'Column') -> bool:
        if not self.is_string() or not other_column.is_string():
            return False

        return other_column.string_size() > self.string_size()

    def _inner_dtype(self, dtype) -> str:
        inner_dtype = dtype.strip()

        if low_card_match := self._low_card_regex.search(inner_dtype):
            self.is_low_cardinality = True
            inner_dtype = low_card_match.group(1)

        if null_match := self._nullable_regex.search(inner_dtype):
            self.is_nullable = True
            inner_dtype = null_match.group(1)

        return inner_dtype


@dataclass(frozen=True)
class ClickHouseColumnChanges:
    on_schema_change: Literal['ignore', 'fail', 'append_new_columns', 'sync_all_columns']
    columns_to_add: List[Column] = field(default_factory=list)
    columns_to_drop: List[Column] = field(default_factory=list)
    columns_to_modify: List[Column] = field(default_factory=list)

    def __bool__(self) -> bool:
        return bool(self.columns_to_add or self.columns_to_drop or self.columns_to_modify)

    @property
    def has_schema_changes(self) -> bool:
        return bool(self)

    @property
    def has_sync_changes(self) -> bool:
        return bool(self.columns_to_drop or self.columns_to_modify)

    @property
    def has_conflicting_changes(self) -> bool:
        if self.on_schema_change == 'fail' and self.has_schema_changes:
            return True

        if self.on_schema_change != 'sync_all_columns' and self.has_sync_changes:
            return True

        return False



#!/usr/bin/env python

import os
import re

from setuptools import find_namespace_packages, setup

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md')) as f:
    long_description = f.read()


# get this from a separate file
def _dbt_clickhouse_version():
    _version_path = os.path.join(this_directory, 'dbt', 'adapters', 'clickhouse', '__version__.py')
    _version_pattern = r'''version\s*=\s*["'](.+)["']'''
    with open(_version_path) as f:
        match = re.search(_version_pattern, f.read().strip())
        if match is None:
            raise ValueError(f'invalid version at {_version_path}')
        return match.group(1)


package_name = 'dbt-clickhouse'
package_version = _dbt_clickhouse_version()
description = '''The Clickhouse plugin for dbt (data build tool)'''

dbt_version = '1.8.0'
dbt_minor = '.'.join(dbt_version.split('.')[0:2])

if not package_version.startswith(dbt_minor):
    raise ValueError(
        f'Invalid setup.py: package_version={package_version} must start with '
        f'dbt_version={dbt_minor}'
    )


setup(
    name=package_name,
    version=package_version,
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='ClickHouse Inc.',
    author_email='guy@clickhouse.com',
    url='https://github.com/ClickHouse/dbt-clickhouse',
    license='MIT',
    packages=find_namespace_packages(include=['dbt', 'dbt.*']),
    package_data={
        'dbt': [
            'include/clickhouse/dbt_project.yml',
            'include/clickhouse/macros/*.sql',
            'include/clickhouse/macros/**/*.sql',
        ]
    },
    install_requires=[
        f'dbt-core~={dbt_version}',
        'clickhouse-connect>=0.6.22',
        'clickhouse-driver>=0.2.6',
        'setuptools>=0.69',
    ],
    python_requires=">=3.8",
    platforms='any',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
)
